/* DSP LABORATORY 
 * University of Technology Graz, 2018
 * Josef Kulmer
 *
 * Obtained from 
 * V1.0  Oct.02, 2005,  C.Feldbauer <feldbauer@tugraz.at>
 */

// interface for channel-wise sample-by-sample processing
#include "Quam.h"
#include "blockprocessing.h"

Quam::Quam() {
    // constructor
    sync = 0.9; // amplitude of the sync signal 
    A = 0.3; // Symbol amplitude
    counter = 0; // counter of output samples (samples per symbol)
    n = 0; // time index for the carrier generation
    CHn = 0; // time index for the time varying channel
    pi = 3.14159265358979;
    curr_I = -1;
    curr_Q = -1; // pointers to the head of the FIR delay lines 
    I = 0;
    Q = 0; // inphase and quadrature 
    OutI = 0;
    OutQ = 0;
    SinCarrier = 0;
    CosCarrier = 0; // carriers

    int i = 1;
    // clearing the delays for the FIR filters
    for (i = 0; i < (2 * N - 1); i++) {
        z_I[i] = 0;
        z_Q[i] = 0;
    }
};

Quam::~Quam() {
};

// This function implements FIR filter
// function interface
//   int *curr : a pointer to the head of the circular delay line 
//   float input : current input sample
//   int ntaps  : number of taps in the filter
//   float h[] : filter IR
//   float z[] : delay line of the filter
float Quam::fir(int *curr, float input, int ntaps, float h[], float z[]) {
    int i = 0, curr_i = 0;
    float Rez = 0;

    // circularly increment the pointer to the head 
    (*curr)++;
    // check if we have reached the end of the delay line
    if (*curr == ntaps)
        *curr = 0;
    // put a new sample in the delay line 
    z[*curr] = input;

    // implement the convolution sum
    curr_i = *curr;
    Rez = 0;
    for (i = 0; i < ntaps; i++) {
        // here we read the elements from the delay line backwards
        Rez += h[i] * z[curr_i];
        // decrement the index of the input circularly
        --curr_i;
        // check if the beginning of the delay line is reached
        if (curr_i == -1)
            curr_i = ntaps - 1;
    }

    return Rez;
}

// QAM mapping of the bits into symbols

void Quam::QAMmap() {
    int i = 0;

    I = 0;
    Q = 0;
    for (i = 0; i < POW; i++) {
        I = (float) b[2 * i] + I * 0.5;
        Q = (float) b[2 * i + 1] + Q * 0.5;
    }
    I = I*A;
    Q = Q*A;
}


// the very simple channel (c+jd)delta[n];
// i.e., a constant (complex) multiplier

void Quam::channel() {
    // real and imaginary part of the channel 
    float c = 0, d = 0;

    // very simple channel
    if (INC_TRANS_CHANNEL == 1) {
        c = 1.1;
        d = -0.45;
    }
    // time-varying channel
    if (INC_TRANS_CHANNEL == 2) {
        // channel varies at a rate of 1Hz
        // only the phase is affected!!!!
        c = 1 * cosf(2 * pi * CHn * 1 / FS);
        d = 1 * sinf(2 * pi * CHn * 1 / FS);
    }
    CHn++;
    if (CHn == FS)
        CHn = 0;

    I = I * c - Q*d;
    Q = I * d + Q*c;
}

// 4-QAM modulation

void Quam::QAMmod() {
    int i = 0;
    // generate random bits
    for (i = 0; i < NBits; i++)
        b[i] = ((rand() % 2) > 0) ? 1 : -1;
    
    // Mapping the bits to QAM 
    QAMmap();
    // include channel if needed 
#if INC_TRANS_CHANNEL!=0
    channel();
#endif 
}

// carrier generation

void Quam::carrier() {
    CosCarrier = 2 * cosf(2 * pi * CARRIERFREQ * n / FS);
    SinCarrier = 2 * sinf(2 * pi * CARRIERFREQ * n / FS);
    n++;
    if (n == FS)
        n = 0;

}

/***********************************************/
/**  Modulator                                 */

/***********************************************/

void Quam::qam() {

    // carrier generation 
#if MODULATE==1
    carrier();
#endif


    if (counter < N) {
        // wait for the next symbol period


        //Output Quadrature and In-phase components 
        //  PUT YOUR MODIFICATIONS HERE -----
        OutI = 0;
        OutQ = 0;
        //-----------------------------------


        counter++;
    } else {
        // The new symbol arrives
        // QAM pulse modulator
        QAMmod();


        //Output Quadrature and In-phase components 
        //  PUT YOUR MODIFICATIONS HERE  ---------
        OutI = (float) I;
        OutQ = (float) Q;
        //----------------------------------------

        // set counter two 1, since one sample has been 
        // has been already sent to the output 
        counter = 1;

        // inverting the sync sign on the symbol boundary
        sync = -1 * sync;
    }

#if MODULATE==1
    OutI = OutI*CosCarrier;
    OutQ = OutQ*SinCarrier;
#endif
#if SINGLECHANNEL==1
    // reducing the Vpp of the resulting 
    // waveform
    OutQ = 0.5 * (OutQ - OutI);
    // the other channel is used for sync
    OutI = sync;
#endif 
}

// output channels

float Quam::processing_l() {
    return OutI;
}

float Quam::processing_r() {
    return OutQ;
}
